/*
 * btype.cpp - part of btype
 * Copyright (C) 2005 Michael Pyne <michael.pyne@kdemail.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <qiodevice.h>
#include <qtextstream.h>
#include <qstring.h>
#include <qbuffer.h>

#include <kdebug.h>

#include "btype.h"
#include "bytetape.h"
#include "sha1.h"
#include "sha1hash.h"

// Implemented for QMap.  Does a simple lexicographical comparison.
bool operator<(const QByteArray &l, const QByteArray &r)
{
    unsigned minLen = QMIN(l.size(), r.size());

    for(unsigned i = 0; i < minLen; ++i) {
        if(l[i] < r[i])
            return true;
        if(l[i] > r[i])
            return false;
    }

    // If we get here they both have a common prefix or something.
    if(l.size() < r.size())
        return true;

    return false;
}

// Since we used pointers in our collections, we need to make sure they're freed
// when we get destroyed.  Luckily KSharedPtr should take care of that for us.
BType::~BType()
{
}

// Construct a Null device.  This will likely throw if you ever use it again.
BType::BType() : KShared(), m_type(Null), m_int(0)
{
}

// Copy ctor.
BType::BType(const BType &other) : KShared(other), m_type(other.m_type), m_data(other.m_data),
                                   m_list(other.m_list), m_dict(other.m_dict),
                                   m_int(other.m_int)
{
}

// Construct an Int.
BType::BType(int i) : KShared(), m_type(Int), m_int(i)
{
}

// Construct a String.  Assumes that it's already encoded as utf8.
BType::BType(const QCString &s) : KShared(), m_type(String), m_data(s), m_int(0)
{
}

// Construct a String.  Uses utf8.
BType::BType(const QString &s) : KShared(), m_type(String), m_data(s.utf8()), m_int(0)
{
}

// Constructs a String.  Note that a null-terminator will be added whether or
// not one was already present (since it is removed by bEncode()).
BType::BType(const QByteArray &data) : KShared(), m_type(String), m_data(data), m_int(0)
{
    m_data.resize(m_data.size() + 1);
    m_data[m_data.size() - 1] = 0;
}

// Constructs a List.
BType::BType(const BList &list) : KShared(), m_type(List), m_list(list), m_int(0)
{
}

// Constructs a Dict.
BType::BType(const BDict &dict) : KShared(), m_type(Dict), m_dict(dict), m_int(0)
{
}

// Constructs a BType based on the input from @p in.  This ctor will
// most likely throw on any kind of error.
BType::BType(QIODevice &in) : KShared(), m_type(Null), m_int(0)
{
    QByteArray data(in.readAll());

    // The ByteTape class is a subclass of QByteArray with the notion of a
    // current pointer.
    ByteTape bt(data);
    BType ret = readType(bt);

    if(!bt.eof())
        kdWarning() << "ByteTape not at EOF after done reading from it.\n";

    *this = ret; // Sucks that we use operator=() in a ctor.
}

// Reads in a type based in the scanned input.
BType BType::readType(ByteTape &in)
{
    if(in.eof())
        throw "Premature EOF in input.";

    switch(*in)
    {
        case 'd': // Dictionary
            return readDict(in);
        case 'i': // Integer
            return readInt(in);
        case 'l': // List
            return readList(in);
    }

    if(QChar(*in).isDigit()) {
        // String
        return readString(in);
    }
    else
        throw "Invalid format";
}

// Reads in a dictionary format.
BType BType::readDict(ByteTape &in)
{
    if(*in != 'd')
        throw "Error trying to read dictionary.";

    ++in; // Skip the 'd'

    BDict dict;

    while(*in != 'e')
    {
        BType str = readString(in);
        BType value = readType(in);

        dict[str.asData()] = new BType(value);
    }

    ++in; // Skip 'e'
    return BType(dict);
}

// Reads in an integer.
BType BType::readInt(ByteTape &in)
{
    if(*in != 'i')
        throw "Error trying to read integer.";

    ++in; // Skip the 'i'

    // Note the the ByteTape functions generally work relative to the current
    // position.  For example, that's what happens with find, it returns the
    // relative position to the entry searched for.
    int ePos = in.find('e');
    if(ePos == -1)
        throw "Unterminated integer";

    // slice() reads starting from the current position up to, but not including,
    // the given offset.
    QByteArray digits = in.slice(ePos);

    in += ePos + 1; // Skip digits and 'e'

    bool ok = true;
    int value = QString(digits).toInt(&ok);

    if(!ok)
        throw "Invalid integer";

    return BType(value);
}

// Reads in a string.  The notion of a string is weird with b-encoding, it can
// be any kind of binary data.  If it is textual, it will be formatted in
// UTF-8.
BType BType::readString(ByteTape &in)
{
    int colon = in.find(':');

    if(colon == -1)
        throw "Couldn't find : in String";

    QByteArray num = in.slice(colon);
    bool ok;
    int size = QString(num).toInt(&ok);

    if(!ok || size < 0) {
        num[num.size() - 1] = 0;
        kdError() << "Invalid number " << QString(num) << " at " << in.pos() << endl;
        throw "Invalid number reading in string.";
    }

    in += colon + 1; // Skip digits and ':'

    QByteArray str;
    if(size > 0) {
        str = in.slice(size);

        if((int) str.size() != size)
            throw "String too short, perhaps the file is corrupt?";
    }

    // Advance past string.
    in += size;

    return BType(str);
}

// Reads in a list.
BType BType::readList(ByteTape &in)
{
    if(*in != 'l')
        throw "Error trying to read list.";

    ++in; // Skip the 'l'
    BList list;

    while(*in != 'e')
        list.append(new BType(readType(in)));

    ++in; // Skip 'e'
    return BType(list);
}

// Returns the number of items in either a Dict or List.
unsigned BType::count() const
{
    if(m_type == List)
        return m_list.count();

    if(m_type == Dict)
        return m_dict.count();

    throw "Unsupported operation";
}

// Assignment operator.
BType &BType::operator=(const BType &other)
{
    m_type = other.m_type;
    m_data = other.m_data;
    m_list = other.m_list;
    m_dict = other.m_dict;
    m_int = other.m_int;

    KShared::operator=(other);

    return *this;
}

SHA1Hash BType::sha1sum() const
{
    QBuffer dataBuf;
    dataBuf.open(IO_WriteOnly);

    bEncode(dataBuf);

    SHA1 sha1;
    sha1.process(dataBuf.buffer().data(), dataBuf.buffer().size());

    return SHA1Hash(sha1);
}

// Output a nice, more human-readable representation of the data.  Note that
// this isn't b-encoded.  See bEncode() for that.
QTextStream &operator<<(QTextStream &out, const BType &data)
{
    BTypeIterator it;

    switch(data.type()) {
        case BType::Null:
            out << "null";
            break;

        case BType::Int:
            out << data.asInt();
            break;

        case BType::String:
            out << '\'' << data.asString() << '\'';
            break;

        case BType::List:
            it = data.begin();
            if(it != data.end())
                out << "[ ";
            else {
                // No data in list.
                out << "[ ]";
                break;
            }

            while(it != data.end()) {
                out << *it;

                ++it;
                if(it != data.end())
                    out << ", ";
            }

            out << " ]";

            break;

        case BType::Dict:
            it = data.begin();
            if(it != data.end())
                out << "{ ";
            else {
                // No data in dict.
                out << "{ }";
                break;
            }

            while(it != data.end()) {
                out << '\'' << QString::fromUtf8(it.key()) << "' => " << *it;

                ++it;
                if(it != data.end())
                    out << ", ";
            }

            out << " }";

            break;
    }

    return out;
}

// Returns a simple string containing the typename of the object.  Use
// type() if you want to compare types, it's much faster.
QString BType::typeName() const
{
    switch(m_type)
    {
        case Int:
            return "Int";
        case String:
            return "String";
        case List:
            return "List";
        case Dict:
            return "Dict";
        default:
            return "Null";
    }
}

// Returns the QString form of the String data.  Weird things will happen if
// the String data is in some kind of binary format.
QString BType::asString() const
{
    if(m_type != String)
        throw "Can't call asString on an object that isn't a String!";

    return QString::fromUtf8(m_data.data(), m_data.size());
}

// Returns the integer value of Int objects.
int BType::asInt() const
{
    if(m_type != Int)
        throw "Can't call asInt on an object that isn't an Int!";

    return m_int;
}

// Outputs the object as a b-encoded representation of the data.
// The dictionary keys are sorted before writing them out, because you should
// always get back exactly the same b-encoded data you read in when you write
// it back out.
void BType::bEncode(QIODevice &out) const
{
    unsigned len;

    switch(m_type) {
        case Dict:
        {
            QTextStream o(&out);
            o << 'd';
            for(BDict::ConstIterator it = m_dict.begin(); it != m_dict.end(); ++it)
            {
                len = it.key().size() - 1;
                o << len << ':';
                out.writeBlock(it.key().data(), len);

                (*it)->bEncode(out);
            }
            o << 'e';
        }
        break;

        case Int:
        {
            QTextStream o(&out);
            o << 'i' << m_int << 'e';
        }
        break;

        case String:
        {
            QTextStream o(&out);

            len = m_data.size() - 1;
            o << len << ':';
            out.writeBlock(m_data.data(), len);
        }
        break;

        case List:
        {
            out.putch('l');
            for(BList::ConstIterator it = m_list.begin(); it != m_list.end(); ++it)
                (*it)->bEncode(out);

            out.putch('e');
        }
        break;

        default:
            throw "Invalid case in BType::bEncode().";
    }
}

// Returns the raw data representation of the String object.
DictKey BType::asData() const
{
    if(m_type != String)
        throw "Can't call asString on an object that isn't a String!";

    return m_data;
}

// Dereference a Dict.
BType &BType::operator[](const DictKey &key)
{
    if(m_type != Dict)
        throw "Can't dereference an object that's not a Dictionary!";

    if(!m_dict.contains(key))
        throw "Unknown key";

    return *m_dict[key];
}

// Dereference a Dict.
const BType &BType::operator[](const DictKey &key) const
{
    if(m_type != Dict)
        throw "Can't dereference an object that's not a Dictionary!";

    if(!m_dict.contains(key))
        throw "Unknown key";

    return *m_dict[key];
}

BType &BType::operator[](const BType &key)
{
    return operator[](key.asData());
}

const BType &BType::operator[](const BType &key) const
{
    return operator[](key.asData());
}

// Returns true if the Dictionary has a given entry.
bool BType::has(const DictKey &key) const
{
    if(m_type != Dict)
        throw "Don't call dictionary functions on non-dictionaries.";

    return m_dict.contains(key);
}

BType &BType::operator[](int i)
{
    if(m_type != List)
        throw "Can't dereference an object that's not a List!";

    if(i < 0 || i >= (int) m_list.count())
        throw "Invalid list index";

    return *m_list[i];
}

const BType &BType::operator[](int i) const
{
    if(m_type != List)
        throw "Can't dereference an object that's not a List!";

    if(i < 0 || i >= (int) m_list.count())
        throw "Invalid list index";

    return *m_list[i];
}

BType::Iterator BType::begin()
{
    if(m_type != List && m_type != Dict)
        throw "No iterator for this kind of BType!";

    if(m_type == List)
        return BTypeIterator(m_list.begin());
    else
        return BTypeIterator(m_dict.begin());
}

BType::Iterator BType::end()
{
    if(m_type != List && m_type != Dict)
        throw "No iterator for this kind of BType!";

    if(m_type == List)
        return BTypeIterator(m_list.end());
    else
        return BTypeIterator(m_dict.end());
}

// The const_cast<> stuff is evil.  This needs to be rewritten to have a
// dedicated ConstIterator type.
BType::ConstIterator BType::begin() const
{
    if(m_type != List && m_type != Dict)
        throw "No iterator for this kind of BType!";

    if(m_type == List)
        return BTypeIterator(const_cast<BType *>(this)->m_list.begin());
    else
        return BTypeIterator(const_cast<BType *>(this)->m_dict.begin());
}

BType::ConstIterator BType::end() const
{
    if(m_type != List && m_type != Dict)
        throw "No iterator for this kind of BType!";

    if(m_type == List)
        return BTypeIterator(const_cast<BType *>(this)->m_list.end());
    else
        return BTypeIterator(const_cast<BType *>(this)->m_dict.end());
}

void BType::append(const BTypePtr data)
{
    if(m_type != List)
        throw "Unsupported operation";

    m_list.append(data);
}

void BType::prepend(const BTypePtr data)
{
    if(m_type != List)
        throw "Unsupported operation";

    m_list.prepend(data);
}

void BType::insert(Iterator &it, const BType &data)
{
    if(m_type != List)
        throw "Unsupported operation";

    m_list.insert(it.m_listIt, new BType(data));
}

void BType::insert(const DictKey &key, BTypePtr value)
{
    if(m_type != Dict)
        throw "Unsupported operation";

    m_dict[key] = value;
}

BTypeIterator BType::find(BTypePtr item)
{
    if(m_type == List)
        return BTypeIterator(m_list.find(item));

    if(m_type == Dict) {
        BDict::Iterator it = m_dict.begin();
        for(; it != m_dict.end(); ++it)
            if(*it == item)
                return BTypeIterator(it);

        return BTypeIterator(m_dict.end());
    }

    throw "Unsupported operation";
}

void BType::remove(Iterator &it)
{
    if(m_type == Dict)
        m_dict.remove(it.m_dictIt);
    else if(m_type == List)
        m_list.remove(it.m_listIt);
    else
        throw "Unsupported operation";
}

void BType::remove(BTypePtr item)
{
    BTypeIterator it = find(item);
    remove(it);
}

//
// Implementation of BTypeIterator.
//

BTypeIterator::BTypeIterator() : m_type(BType::Null)
{
}

BTypeIterator::BTypeIterator(BDict::Iterator it) : m_dictIt(it), m_type(BType::Dict)
{
}

BTypeIterator::BTypeIterator(BList::Iterator it) : m_listIt(it), m_type(BType::List)
{
}

BTypeIterator::BTypeIterator(const BTypeIterator &other) : m_dictIt(other.m_dictIt),
                                                           m_listIt(other.m_listIt),
                                                           m_type(other.m_type)
{
}

BTypeIterator &BTypeIterator::operator=(const BTypeIterator &other)
{
    m_dictIt = other.m_dictIt;
    m_listIt = other.m_listIt;
    m_type = other.m_type;

    return *this;
}

const BType &BTypeIterator::operator*() const
{
    if(m_type == BType::List)
        return *(*m_listIt);
    else if(m_type == BType::Dict)
        return *(*m_dictIt);
    else
        throw "Invalid dereference of null iterator!";
}

BTypeIterator::operator BTypePtr()
{
    if(m_type == BType::List)
        return *m_listIt;
    else if(m_type == BType::Dict)
        return *m_dictIt;

    throw "Invalid dereference of null iterator!";
}

BTypeIterator::operator BTypePtr() const
{
    if(m_type == BType::List)
        return *m_listIt;
    else if(m_type == BType::Dict)
        return *m_dictIt;

    throw "Invalid dereference of null iterator!";
}

BType &BTypeIterator::operator*()
{
    if(m_type == BType::List)
        return *(*m_listIt);
    else if(m_type == BType::Dict)
        return *(*m_dictIt);
    else
        throw "Invalid dereference of null iterator!";
}

BTypeIterator &BTypeIterator::operator++()
{
    if(m_type == BType::Null)
        throw "Can't advance null iterator!";

    if(m_type == BType::List)
        ++m_listIt;
    else
        ++m_dictIt;

    return *this;
}

BTypeIterator &BTypeIterator::operator--()
{
    if(m_type == BType::Null)
        throw "Can't decrement null iterator!";

    if(m_type == BType::List)
        --m_listIt;
    else
        --m_dictIt;

    return *this;
}

bool BTypeIterator::operator==(const BTypeIterator &other) const
{
    if(m_type == BType::Null || other.m_type == BType::Null)
        throw "Can't compare null iterators with other types.";

    if(m_type != other.m_type)
        throw "Can't compare list and dictionary iterators!";

    if(m_type == BType::List)
        return m_listIt == other.m_listIt;
    else
        return m_dictIt == other.m_dictIt;
}

const QByteArray &BTypeIterator::key() const
{
    if(m_type != BType::Dict)
        throw "Only dictionaries have keys!";

    return m_dictIt.key();
}

bool BTypeIterator::operator!=(const BTypeIterator &other) const
{
    return !(*this == other);
}

// vim: set et ts=8 sw=4:
