/*
 * btype.h - part of btype
 * Copyright (C) 2005 Michael Pyne <michael.pyne@kdemail.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef BTYPE_H
#define BTYPE_H

#include <ksharedptr.h>

#include <qcstring.h>
#include <qvaluelist.h>
#include <qmap.h>
#include <qstring.h>

class QIODevice;
class QTextStream;

class BType;
class BTypeIterator;
class ByteTape;

typedef KSharedPtr<BType> BTypePtr;

class SHA1Hash;

// Defined for QMap
bool operator<(const QByteArray &l, const QByteArray &r);

class DictKey : public QByteArray
{
public:
    DictKey() : QByteArray() { }
    DictKey(const QByteArray &other) : QByteArray(other) { }
    DictKey(const QString &other) : QByteArray(other.utf8()) { }
};

// Typedef for list and maps of our type.  We have to use pointers because
// otherwise there is infinite recursion.
typedef QValueList<BTypePtr> BList;
typedef QMap<DictKey, BTypePtr> BDict;

/**
 * A polymorphic type to represent the possible values of a b-encoded file.
 * Can represent lists, dictionaries, and the int and string basic types.
 *
 * Some may find this class evil... I like to think of it as underappreciated
 * =)
 */
class BType : public KShared
{
public:
    /// The various forms we can assume.
    enum Type { Null, String, Int, List, Dict };

    /// Iterator declarations.  Both types are of BTypeIterator because I am
    /// too lazy to write BConstTypeIterator.
    typedef BTypeIterator Iterator;
    typedef BTypeIterator ConstIterator;

    ~BType();

    BType(); ///< Construct a null object.

    // Copy ctor.
    BType(const BType &other);

    // Construct an Int
    BType(int i);

    // Construct a String.
    BType(const QCString &s);
    BType(const QString &s);

    // This one accepts arbitrary data.
    BType(const QByteArray &data);

    // Construct a list
    BType(const BList &list);

    // Construct a dict
    BType(const BDict &dict);

    // The ctor most people will use in one form or another.  Notice that this
    // uses the assignment operator internally, but the default ctor is more
    // or less quick so that should be ok.
    BType(QIODevice &in);

    BType &operator=(const BType &other);

    // Various introspection functions.
    Type type() const { return m_type; }
    QString typeName() const;

    bool isNull() const { return m_type != Null; }
    bool isList() const { return m_type == List; }
    bool isString() const { return m_type == String; }
    bool isDict() const { return m_type == Dict; }
    bool isInt() const { return m_type == Int; }

    // Returns the 20-byte SHA1 sum of this object's bencoded data (including
    // any children for lists and dicts).
    SHA1Hash sha1sum() const;

    // If, and only if, this object is a Dict, returns whether or not a given
    // entry is in the dictionary.  Entries are keyed by QByteArray since the
    // underlying b-encoded string type supports arbitrary data.  Most dict
    // functions accept various overloads to make them easier to use.
    bool has(const DictKey &key) const;

    // String return functions.  asString() returns a QString, and is
    // therefore inappropriate for entries that may contain binary data.
    // The encoding used is UTF-8, the standard for b-encoded data.
    QString asString() const;
    DictKey asData() const;

    // Returns the integer value iff this is an Int.
    int asInt() const;

    // Output a b-encoded form of this object to the given QIOObject.
    void bEncode(QIODevice &out) const;

    // Dereference Dict-type objects.
    BType &operator[](const DictKey &key);
    const BType &operator[](const DictKey &key) const;

    // This works iff key is a String, otherwise an exception is thrown.
    BType &operator[](const BType &key);
    const BType &operator[](const BType &key) const;

    // List dereference operator.
    BType &operator[](int i);
    const BType &operator[](int i) const;

    // Add to end of list.
    void append(const BTypePtr data);

    // Add to beginning of list.
    void prepend(const BTypePtr data);

    // Insert before the given iterator.
    void insert(Iterator &it, const BType &data);

    // Insert into the dictionary.
    void insert(const DictKey &key, BTypePtr value);

    // Tries to find the given item, returns end() if there's no match.
    Iterator find(BTypePtr item);

    // Removes the item pointed at by it.
    void remove(Iterator &it);

    // Removes all items equal to the given item.
    void remove(BTypePtr item);

    // Returns the count of entries for Lists and Dicts.
    unsigned count() const;

    // Returns a dual-mode iterator representing the start of either a list
    // or dict, depending on type().
    Iterator begin();
    Iterator end();

    ConstIterator begin() const;
    ConstIterator end() const;

private:
    // Internal functions to read in data from the ByteTape utility class.
    BType readDict(ByteTape &in);
    BType readInt(ByteTape &in);
    BType readString(ByteTape &in);
    BType readList(ByteTape &in);

    // Scans the input to determine the type before passing it onto one of the
    // other read*() functions.
    BType readType(ByteTape &in);

private:
    Type m_type;
    QByteArray m_data;
    BList m_list;
    BDict m_dict;
    int m_int;
};

// Dual-mode iterator for BType.  Can be used to iterate over BType::List and
// BType::Dict.  Some functions only work for BType::Dict.
class BTypeIterator
{
    friend class BType;

public:
    BTypeIterator();
    BTypeIterator(BDict::Iterator it);
    BTypeIterator(BList::Iterator it);
    BTypeIterator(const BTypeIterator &other);

    BTypeIterator &operator=(const BTypeIterator &other);

    const BType &operator*() const;
    BType &operator*();

    // Only works for BType::Dict
    const QByteArray &key() const;

    BTypeIterator &operator++();
    BTypeIterator &operator--();

    // These will throw exceptions if the iterators are of different types
    // instead of just returning false.
    bool operator ==(const BTypeIterator &other) const;
    bool operator !=(const BTypeIterator &other) const;

    operator BTypePtr();
    operator BTypePtr() const;

    // Returns the type of the iterator.
    BType::Type type() const { return m_type; }

private:
    BDict::Iterator m_dictIt;
    BList::Iterator m_listIt;

    BType::Type m_type;
};

// Outputs a decent textual representation of the data to out.  No attempt is
// made to encode binary data so it's safe for terminals.
QTextStream &operator<<(QTextStream &out, const BType &data);

#endif /* BTYPE_H */

// vim: set et ts=8 sw=4:
