/*
 * bytetape.cpp - part of btype
 * Copyright (C) 2005 Michael Pyne <michael.pyne@kdemail.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <kdebug.h>

#include "bytetape.h"

ByteTape::ByteTape() : QByteArray(), m_data(new ByteTapeData)
{
}

ByteTape::~ByteTape()
{
}

ByteTape::ByteTape(const QByteArray &data, int pos) : QByteArray(data),
						      m_data(new ByteTapeData(pos))
{
}

ByteTape::ByteTape(ByteTape &other) : QByteArray(other), m_data(other.m_data)
{
}

ByteTape &ByteTape::operator=(ByteTape &other)
{
    m_data = other.m_data;
    QByteArray::operator=(other);

    return *this;
}

char ByteTape::operator*() const
{
    if(eof()) {
	kdWarning() << "Dereferencing EOF\n";
	return 0;
    }

    return at(m_data->pos);
}

ByteTape &ByteTape::operator++()
{
    if(eof()) {
	kdWarning() << "Trying to seek past EOF\n";
	return *this;
    }

    ++m_data->pos;
    return *this;
}

ByteTape &ByteTape::operator--()
{
    if(m_data->pos <= 0) {
	kdWarning() << "Trying to set negative size.\n";
	return *this;
    }

    --m_data->pos;
    return *this;
}

ByteTape &ByteTape::operator+=(int i)
{
    if(m_data->pos + i > (int) size()) {
	kdWarning() << "Trying to seek past EOF\n";
	m_data->pos = size(); // eof
    }

    m_data->pos += i;
    return *this;
}

ByteTape &ByteTape::operator-=(int i)
{
    if(m_data->pos - i < 0) {
	kdWarning() << "Trying to seek beyond begin of stream.\n";
	m_data->pos = 0;
	return *this;
    }

    m_data->pos -= i;
    return *this;
}

int ByteTape::find(char ch) const
{
    int absolutePosition = QByteArray::find(ch, m_data->pos);
    if(absolutePosition == -1)
	return -1;

    return absolutePosition - m_data->pos;
}

void ByteTape::setPos(int pos)
{
    if(pos < 0) {
	kdWarning() << "Trying to seek beyond begin of stream.\n";
	pos = 0;
    }

    if(pos > (int) size()) {
	kdWarning() << "Trying to seek past EOF.\n";
	pos = size(); // eof
    }

    m_data->pos = pos;
}

// This slice is [pos, pos + size).  In other words, it includes the current
// position, and excludes the endpoint.
QByteArray ByteTape::slice(int size) const
{
    if(size == 0)
	return QByteArray();

    if(size < 0) {
	kdWarning() << "Trying to grab negative-sized memory block.\n";
	return QByteArray();
    }

    QByteArray retData(size);
    qmemmove(retData.data(), &at(m_data->pos), size);

    return retData;
}

bool ByteTape::eof() const
{
    return m_data->pos == (int) size();
}

