/*
 * mainform.cpp - part of btype
 * Copyright (C) 2005 Michael Pyne <michael.pyne@kdemail.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <kfiledialog.h>
#include <kstdaction.h>
#include <kinputdialog.h>
#include <kpopupmenu.h>
#include <kaction.h>
#include <kdebug.h>
#include <klocale.h>
#include <kmessagebox.h>

#include <qpainter.h>
#include <qfont.h>
#include <qvbox.h>
#include <qfile.h>
#include <qbuffer.h>
#include <qtextstream.h>

#include <limits.h>

#include "mainform.h"
#include "sha1hash.h"

void debug(const BType &obj)
{
    QBuffer buf;
    buf.open(IO_WriteOnly);
    QTextStream ts(&buf);

    ts << obj;

    kdDebug() << "Data: " << QString(buf.buffer()) << endl;
}

void BTypeListViewItem::paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align)
{
    QColorGroup newCg(cg);

    // The 3rd column is the value column, 1st is name (if any), 2nd is
    // the type (which should always be String for this class).
    if(column == 2 && m_binaryString) {
        QFont f = p->font();
        f.setItalic(true);
        p->setFont(f);

        newCg.setColor(QColorGroup::Text, newCg.link());
    }

    KListViewItem::paintCell(p, newCg, column, width, align);
}

void BTypeListViewItem::checkForBinaryString()
{
    m_binaryString = false;
    if(!m_obj->isString())
        return;

    QString s = m_obj->asString();

    m_binaryString = s.length() > 300;
    unsigned i = 0;

    while(!m_binaryString && i < s.length())
        m_binaryString = !s[i++].isPrint();

    if(m_binaryString)
        setText(2, i18n("Binary data"));
}

/**
 * Hide the data value with a simple "Binary Data" blurb, and use an italic
 * font as well.  Also make it a different color while we're going for the
 * iCandy award.
 */
BTypeListViewItem *createNewItem(QListViewItem *parent, QListView *listViewParent,
        const QString &name, const QString &type, const QString &data,
        BTypePtr obj)
{
    if(parent)
        return new BTypeListViewItem(parent, name, type, data, obj);
    else
        return new BTypeListViewItem(listViewParent, name, type, data, obj);
}

BTypeListViewItem::BTypeListViewItem(QListViewItem *parent, const QString &name,
        const QString &type, const QString &data, BTypePtr obj)
    : KListViewItem(parent, name, type, data), m_data(data), m_obj(obj),
      m_binaryString(false)
{
    checkRenamability();
    if(m_obj->isString())
        checkForBinaryString();
}

BTypeListViewItem::BTypeListViewItem(QListView *parent, const QString &name,
        const QString &type, const QString &data, BTypePtr obj)
    : KListViewItem(parent, name, type, data), m_data(data), m_obj(obj),
      m_binaryString(false)
{
    checkRenamability();
    if(m_obj->isString())
        checkForBinaryString();
}

void BTypeListViewItem::checkRenamability()
{
    // Check if parent is a dict, if so allow renaming of column 0.
    if(parent() &&
       static_cast<BTypeListViewItem *>(parent())->obj()->isDict())
    {
        setRenameEnabled(0, true);
    }

    // Allow inline renaming for data types.
    if(obj()->isString() || obj()->isInt())
        setRenameEnabled(2, true);
}

MainForm::MainForm(QWidget *parent) : MainFormBase(parent, "btypeAppMainForm"),
                                      m_dirty(false)
{
    newDocument();
}

bool MainForm::queryClose()
{
    if(m_dirty) {
        switch(doSave()) {
            case KMessageBox::Yes:
                if(!save())
                    return false;
            break;

            case KMessageBox::No:
                return true;
            break;

            default:
                return false;
        }
    }

    return true;
}

void MainForm::doInsert(const BTypePtr b, QListViewItem *parent, const QString &name)
{
    // This should be mostly self-explanatory.
    switch(b->type()) {
        case BType::Int:
            createNewItem(parent, m_listView, name, i18n("Integer"), QString::number(b->asInt()), b);
        break;

        case BType::List:
        {
            parent = createNewItem(parent, m_listView, name, i18n("List"), "[ ]", b);

            // We simply use a number as the name of each entry.
            unsigned i = 0;
            int numDigits = numDecimalDigits(b->count());

            for(BType::Iterator it = b->begin(); it != b->end(); ++it)
                doInsert(it, parent, numToString(++i, numDigits));
        }
        break;

        case BType::String:
        {
            createNewItem(parent, m_listView, name, i18n("String"), b->asString(), b);
        }
        break;

        case BType::Dict:
        {
            parent = createNewItem(parent, m_listView, name, i18n("Dict"), "{ }", b);

            for(BType::Iterator it = b->begin(); it != b->end(); ++it)
                doInsert(it, parent, QString(it.key()));
        }
        break;

        default:
            kdError() << "Unexpected case in " << k_funcinfo << endl;
    }
}

void MainForm::addToCollection(BTypeListViewItem *i, const BTypePtr newObj, const QString &value)
{
    bool ok;

    // If i doesn't exist, the user is trying to insert into an empty listview.
    // Since b-encoded files only support one root node, this is only allowed
    // when we don't already have children.
    if(!i) {
        new BTypeListViewItem(m_listView, "/", newObj->typeName(), value, newObj);
        return;
    }

    if(i->obj()->isDict()) {
        QString name = KInputDialog::getText(i18n("Enter a name"),
                i18n("Enter a name for the new value"),
                QString::null, &ok, this);

        if(!ok || name.isEmpty())
            return;

        (*i->obj()).insert(name, newObj);
        new BTypeListViewItem(i, name, newObj->typeName(), value, newObj);
    }
    else {
        BTypePtr b = i->obj();
        b->append(newObj);
        new BTypeListViewItem(i, numToString(b->count(),
                    numDecimalDigits(b->count())), newObj->typeName(), value, newObj);
    }

    i->setOpen(true);
    setDirty(true);
}

void MainForm::newDocument()
{
    if(m_dirty) {
        int res = doSave();
        if(res == KMessageBox::Cancel)
            return;

        if(res == KMessageBox::Yes && !save())
            return;
    }

    m_listView->clear();

    doInsert(new BType(BDict()), 0, "/");

    m_listView->setRootIsDecorated(true);

    m_fileName = "";
    setDirty(false);
    emit opened(i18n("Untitled"));
}

bool MainForm::open()
{
    if(m_dirty) {
        int res = doSave();
        if(res == KMessageBox::Cancel)
            return true; // We successfully did nothing.

        if(res == KMessageBox::Yes && !save())
            return false;
    }

    QString fn = KFileDialog::getOpenFileName(QString::null, i18n("*.torrent|BitTorrent Files"), this);
    if(fn.isEmpty())
        return false;

    return setup(fn);
}

QString MainForm::getFileName()
{
    return KFileDialog::getSaveFileName(QString::null, i18n("*.torrent|BitTorrent Files"), this);
}

void MainForm::setDirty(bool isDirty)
{
    if(isDirty == m_dirty)
        return;

    m_dirty = isDirty;
    emit dirty(isDirty);
}

bool MainForm::saveAs()
{
    QString newName = getFileName();

    if(newName.isEmpty())
        return false;

    m_fileName = newName;
    return save();
}

bool MainForm::save()
{
    if(m_fileName.isEmpty())
        m_fileName = getFileName();

    if(m_fileName.isEmpty())
        return false;

    // Use a sacrifice file.
    QFile f(m_fileName);

    if(!f.open(IO_WriteOnly)) {
        kdError() << "Unable to open output file " << f.name() << endl;
        return false;
    }

    BTypePtr outputObj;

    // Get first child.  It should be the only child on that level.
    BTypeListViewItem *i = static_cast<BTypeListViewItem *>(m_listView->firstChild());
    if(i->nextSibling()) {
        // Let's auto-construct a list for him so at least everything will get
        // saved.
        BList list;
        BTypeListViewItem *temp = i;

        while(temp) {
            list.append(temp->obj());
            temp = static_cast<BTypeListViewItem *>(temp->nextSibling());
        }

        outputObj = new BType(list);
    }
    else
        outputObj = i->obj();

    // Save to disk
    outputObj->bEncode(f);
    setDirty(false);
    return true;
}

bool MainForm::setup(const QString &fileName)
{
    QFile f(fileName);
    if(!f.open(IO_ReadOnly)) {
        kdError() << "Unable to open " << fileName << endl;
        return false;
    }

    m_listView->clear();

    // BType uses exceptions pretty heavily, so it's important to make
    // sure you catch them.
    try {
        BTypePtr t = new BType(f);

        doInsert(t, 0, "/");

        if(m_listView->firstChild())
            m_listView->firstChild()->setOpen(true);

        m_fileName = fileName;
    }
    catch (const char *e)
    {
        kdError() << "Caught exception: " << e << endl;
        return false;
    }

    // Don't change the order of the next two lines, it breaks the title bar
    // for some reason.
    setDirty(false);
    emit opened(m_fileName);

    return true;
}

int MainForm::doSave()
{
    return KMessageBox::questionYesNoCancel(this,
            i18n("Would you like to save your changes?"), QString::null,
            KStdGuiItem::save(), KStdGuiItem::dontSave());
}

void MainForm::dblClicked(QListViewItem *item)
{
    BTypeListViewItem *i = dynamic_cast<BTypeListViewItem *>(item);
    if(i)
        debug(*i->obj());
}

void MainForm::itemClicked(QListViewItem *item)
{
    if(!item) {
        emit newHash("");
        return;
    }

    BTypeListViewItem *i = static_cast<BTypeListViewItem *>(item);
    emit newHash(i->obj()->sha1sum().hex());
}

void MainForm::renamed(QListViewItem *item, int col, const QString &text)
{
    // We have to update the internal object of the item, and possibly its
    // parent in the case of a dict key rename.

    if(!item)
        return;

    BTypeListViewItem *i = static_cast<BTypeListViewItem *>(item);

    if(col == 0) {
        // Dict key rename, we can assume the parent has a dict.  I think. ;)
        BTypeListViewItem *pIt = static_cast<BTypeListViewItem *>(item->parent());

        // Find old item.
        BType::Iterator it = pIt->obj()->find(i->obj());
        if(it == pIt->obj()->end()) {
            kdError() << "Couldn't find an item, and I know it's there somewhere. :(\n";
            return;
        }

        // Store hash to compare.
        SHA1Hash oldHash = pIt->obj()->sha1sum();

        // Insert correct name.
        pIt->obj()->insert(text, it);

        // Remove old name.
        pIt->obj()->remove(it);

        if(oldHash != pIt->obj()->sha1sum())
            setDirty(true);
    }
    else if (col == 2) {
        // Rename of value.  If we rename an int we need to make sure that the
        // new value is also an integer.

        BTypePtr obj = i->obj();
        if(obj->isInt()) {
            bool ok;
            int newValue = text.toInt(&ok);

            if(!ok) {
                kdWarning() << "Don't rename numbers so that they aren't numbers!\n";
                i->setText(2, QString::number(obj->asInt()));
                return;
            }

            if(obj->asInt() != newValue)
                setDirty(true);

            *obj = BType(newValue);
        }
        else {
            // We have a string, just copy the value into its obj.
            if(obj->asString() != text)
                setDirty(true);

            *obj = BType(text);
        }
    }
}

void MainForm::editSelected()
{
    bool ok;
    QString newValue;
    BTypeListViewItem *item = selectedItem();

    if(!item)
        return;

    BTypePtr b = item->obj();
    BTypePtr newObj;

    if(b->isString()) {
        newValue = KInputDialog::getText(i18n("Change value"), 
                i18n("Enter new text:"),
                b->asString(), &ok, this);

        newObj = new BType(newValue);
    }
    else {
        int newIValue = KInputDialog::getInteger(i18n("Change value"),
                i18n("Enter new number:"),
                b->asInt(), INT_MIN, INT_MAX, 1, &ok, this);

        newValue = QString::number(newIValue);
        newObj = new BType(newIValue);
    }

    if(!ok)
        return;

    SHA1Hash oldHash = b->sha1sum();
    *b = *newObj; // Don't switch pointers, do a deep copy.

    if(b->sha1sum() != oldHash) {
        item->setText(2, newValue);
        setDirty(true);
        emit newHash(b->sha1sum().hex());
    }
}

void MainForm::removeSelected()
{
    BTypeListViewItem *i = selectedItem();

    if(!i)
        return;

    // See if the item has a parent, and if the parent is a collection.
    // If so, remove this item from the parent.
    BTypeListViewItem *parent = static_cast<BTypeListViewItem*>(i->parent());
    if(parent && (parent->obj()->isList() || parent->obj()->isDict()))
        parent->obj()->remove(i->obj());

    delete i;

    setDirty(true);
}

void MainForm::addDictToSelected()
{
    BTypeListViewItem *i = selectedItem();
    if(!i && m_listView->firstChild())
        return;

    BTypePtr newObj = new BType(BDict());
    addToCollection(i, newObj, "{ }");
}

void MainForm::addListToSelected()
{
    BTypeListViewItem *i = selectedItem();
    if(!i && m_listView->firstChild())
        return;

    BTypePtr newObj = new BType(BList());
    addToCollection(i, newObj, "[ ]");
}

void MainForm::addIntToSelected()
{
    BTypeListViewItem *i = selectedItem();
    if(!i && m_listView->firstChild())
        return;

    bool ok;
    int val = KInputDialog::getInteger(i18n("Select new Integer value"),
            i18n("Value:"), 0, INT_MIN, INT_MAX, 1, &ok, this);

    if(!ok)
        return;

    BTypePtr newObj = new BType(val);
    addToCollection(i, newObj, QString::number(val));
}

void MainForm::addStringToSelected()
{
    BTypeListViewItem *i = selectedItem();
    if(!i && m_listView->firstChild())
        return;

    bool ok;
    QString text = KInputDialog::getText(i18n("Select new String value"),
            i18n("Value:"), QString::null, &ok, this);

    if(!ok)
        return;

    BTypePtr newObj = new BType(text);
    addToCollection(i, newObj, text);
}

void MainForm::rightClicked(QListViewItem *item, const QPoint &pos)
{
    // If we have absolutely no children, allow the call to proceed.
    if(!item && m_listView->firstChild())
        return;

    BTypeListViewItem *i = static_cast<BTypeListViewItem *>(item);
    KPopupMenu *menu = new KPopupMenu(this);

    if(i)
        menu->insertTitle(i->obj()->typeName());
    else
        menu->insertTitle(i18n("Empty File"));

    int id;
    if(i)
        id = menu->insertItem(i18n("Edit %1").arg(i->obj()->typeName()), this, SLOT(editSelected()));
    else
        id = menu->insertItem(i18n("Edit Type"), this, SLOT(editSelected()));
    

    if(!i || (!i->obj()->isString() && !i->obj()->isInt()))
        menu->setItemEnabled(id, false);

    menu->insertSeparator();

    QPopupMenu *subMenu = new QPopupMenu(this);
    subMenu->insertItem(i18n("Add New Int"), this, SLOT(addIntToSelected()));
    subMenu->insertItem(i18n("Add New String"), this, SLOT(addStringToSelected()));
    subMenu->insertItem(i18n("Add New Dict"), this, SLOT(addDictToSelected()));
    subMenu->insertItem(i18n("Add New List"), this, SLOT(addListToSelected()));

    id = menu->insertItem(i18n("Remove this entry"), this, SLOT(removeSelected()));
    if(!i)
        menu->setItemEnabled(id, false);

    id = menu->insertItem(i18n("Add New..."), subMenu);
    if(i && !i->obj()->isList() && !i->obj()->isDict())
        menu->setItemEnabled(id, false);

    menu->popup(pos);
}

#include "mainform.moc"

// vim: set et ts=8 sw=4:
