/*
 * mainform.h - part of btype
 * Copyright (C) 2005 Michael Pyne <michael.pyne@kdemail.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef BTYPE_MAINFORM_H
#define BTYPE_MAINFORM_H

#include <klistview.h>

#include "mainformbase.h"
#include "btype.h"

class BTypeListViewItem : public KListViewItem
{
public:
    BTypeListViewItem(QListViewItem *parent, const QString &name,
            const QString &type, const QString &data, BTypePtr obj);

    BTypeListViewItem(QListView *parent, const QString &name,
            const QString &type, const QString &data, BTypePtr obj);

    const BTypePtr obj() const { return m_obj; }
    BTypePtr obj() { return m_obj; }

    /**
     * Changes the color and font for the value column before letting the
     * inherited paintCell() do all the real work.
     */
    virtual void paintCell(QPainter *p, const QColorGroup &cg, int column, int width, int align);

private:
    void checkRenamability();
    void checkForBinaryString();

private:
    QString m_data; ///< Holds a copy of the data so we can print it out if needed.
    BTypePtr m_obj;
    bool m_binaryString;
};

class MainForm : public MainFormBase
{
    Q_OBJECT

public:
    MainForm(QWidget *parent = 0);

    bool isDirty() const { return m_dirty; }
    int doSave();

    bool queryClose();

private slots:
    bool open();
    bool save();
    bool saveAs();
    void newDocument();

    // Called when an item is double clicked, and dumps the internal's object's
    // value to the terminal.
    void dblClicked(QListViewItem *item);

    void itemClicked(QListViewItem *item);
    void renamed(QListViewItem *item, int col, const QString &text);

    void editSelected();
    void removeSelected();

    void addIntToSelected();
    void addDictToSelected();
    void addListToSelected();
    void addStringToSelected();

    void rightClicked(QListViewItem *item, const QPoint &pos);

signals:
    void newHash(const QString &hash);
    void opened(const QString &fileName);
    void dirty(bool isDirty);

private:
    QString getFileName();

    void setDirty(bool isDirty);

    bool setup(const QString &fileName);

    // This function calculates the number of decimal digits needed to write
    // out a number.  It is used to format the list entries properly, since
    // they sort all kinds of funky otherwise.
    int numDecimalDigits(unsigned num)
    {
        int numDigits = 1;

        while(num >= 10) {
            ++numDigits;
            num /= 10;
        }

        return numDigits;
    }

    BTypeListViewItem *selectedItem() const
    {
        return static_cast<BTypeListViewItem*>(m_listView->selectedItem());
    }

    // Function to return a formatted string based on num, which will be
    // zero-padded out to numDigits digits.
    QString numToString(int num, int numDigits)
    {
        QString s = QString::number(num);

        if((int) s.length() < numDigits) {
            QString fill;
            fill.fill('0', numDigits - s.length());
            s.prepend(fill);
        }

        return s;
    }

    // Recursive function to create the list view tree.
    void doInsert(const BTypePtr b, QListViewItem *parent = 0, const QString &name = "");

    // Adds newObj into the list view as a child of i, taking care to add to
    // i's own obj() if necessary.
    void addToCollection(BTypeListViewItem *i, const BTypePtr newObj, const QString &value);

private:
    QString m_fileName;
    bool m_dirty;
};

#endif /* BTYPE_MAINFORM_H */

// vim: set et ts=8 sw=4:
